import re
from typing import Sequence, Optional, Iterable
import itertools
import unidecode

_morse_normalize = str.maketrans("·_—–", ".---")

class MorseChar:
	def __init__(self, dashdots: str):
		if dashdots != "/" and any(c not in ("-.") for c in dashdots):
			raise ValueError("dashdots must be / or a pattern of - and .")
		self.dashdots = dashdots

	def decode(self) -> str:
		return MORSE_DECODE_TABLE.get(self, "�")

	def __eq__(self, other):
		return other is not None and self.dashdots == other.dashdots

	def __hash__(self):
		return hash(self.dashdots)

	def __repr__(self):
		return f"MorseChar('{self.dashdots}')"


MORSE_ENCODE_TABLE = {
	" ": MorseChar("/"),
	"A": MorseChar(".-"),
	"B": MorseChar("-..."),
	"C": MorseChar("-.-."),
	"D": MorseChar("-.."),
	"E": MorseChar("."),
	"F": MorseChar("..-."),
	"G": MorseChar("--."),
	"H": MorseChar("...."),
	"I": MorseChar(".."),
	"J": MorseChar(".---"),
	"K": MorseChar("-.-"),
	"L": MorseChar(".-.."),
	"M": MorseChar("--"),
	"N": MorseChar("-."),
	"O": MorseChar("---"),
	"P": MorseChar(".--."),
	"Q": MorseChar("--.-"),
	"R": MorseChar(".-."),
	"S": MorseChar("..."),
	"T": MorseChar("-"),
	"U": MorseChar("..-"),
	"V": MorseChar("...-"),
	"W": MorseChar(".--"),
	"X": MorseChar("-..-"),
	"Y": MorseChar("-.--"),
	"Z": MorseChar("--.."),
	"0": MorseChar("-----"),
	"1": MorseChar(".----"),
	"2": MorseChar("..---"),
	"3": MorseChar("...--"),
	"4": MorseChar("....-"),
	"5": MorseChar("....."),
	"6": MorseChar("-...."),
	"7": MorseChar("--..."),
	"8": MorseChar("---.."),
	"9": MorseChar("----."),
	".": MorseChar(".-.-.-"),
	",": MorseChar("--..--"),
	"?": MorseChar("..--.."),
	"!": MorseChar("-.-.--"),
	"-": MorseChar("-....-"),
	"/": MorseChar("-..-."),
	":": MorseChar("---..."),
	"'": MorseChar(".----."),
	")": MorseChar("-.--.-"),
	";": MorseChar("-.-.-"),
	"(": MorseChar("-.--."),
	"=": MorseChar("-...-"),
	"@": MorseChar(".--.-."),
	"&": MorseChar(".-..."),
	"+": MorseChar(".-.-."),
	"_": MorseChar("..--.-"),
	'"': MorseChar(".-..-."),
	"$": MorseChar("...-..-"),

	# Prosigns, see https://en.wikipedia.org/wiki/Prosigns-for-MorseChar-code
	"\n": MorseChar(".-.-"),
	"⌫":  MorseChar("........"),
	"🆘": MorseChar("...---..."),
	"⏹":  MorseChar(".-.-."),
	"⏸":  MorseChar(".-..."),
	"✔":  MorseChar("...-."),
	"¿":  MorseChar("..-.-"),
	"🆕": MorseChar("..-.-"),
}

MORSE_DECODE_TABLE = {morse: txt for txt, morse in MORSE_ENCODE_TABLE.items()}


def encode_char(char: str) -> Iterable[Optional[MorseChar]]:
	try:
		yield MORSE_ENCODE_TABLE[char.upper()]
		return
	except KeyError:
		pass

	unidecoded = unidecode.unidecode(char).upper()
	if not unidecoded:
		yield None
		return

	yield from map(MORSE_ENCODE_TABLE.get, unidecoded)

def encode(txt: str) -> Sequence[Optional[MorseChar]]:
	return list(itertools.chain.from_iterable(map(encode_char, txt)))


def decode(morse: Sequence[MorseChar]) -> str:
	return "".join(x.decode() for x in morse)


def dashdots_normalize(dashdots: str) -> str:
	return re.sub(
		r"  *|\n", " ",
		dashdots.translate(_morse_normalize).replace("/", " / ").strip(" ")
	)

def dashdots_to_morsechars(dashdots: str) -> Sequence[MorseChar]:
	return [MorseChar(x) for x in dashdots_normalize(dashdots).split(" ")]

def morsechars_to_dashdots(morse: Sequence[MorseChar]) -> str:
	return " ".join(x.dashdots for x in morse)
